#include <thd_mapping/grid_map.h>

namespace thd_mapping
{

GridMap::GridMap(std::vector<float> &&grid, glm::ivec2 dim, Position origin, float resolution) : grid(std::move(grid)), dim(dim), origin(origin), resolution(resolution), _cov(false)
{
}

GridMap GridMap::downSample(int ratio)
{
    glm::ivec2 newDim = dim / ratio;
    GridMap newGridMap = GridMap(std::vector<float>(newDim.x * newDim.y), newDim, origin, resolution * ratio);
    glm::ivec2 nxy;
    for (nxy.y = 0; nxy.y < newDim.y; nxy.y++)
    {
        for (nxy.x = 0; nxy.x < newDim.x; nxy.x++)
        {
            glm::ivec2 innerXY;
            float value = std::numeric_limits<float>::min();
            for (innerXY.y = 0; innerXY.y < ratio; innerXY.y++)
            {
                for (innerXY.x = 0; innerXY.x < ratio; innerXY.x++)
                {
                    // value = std::min(value,(*this)[nxy * ratio + innerXY]);
                    float& gridValue = (*this)[nxy * ratio + innerXY];
                    value = gridValue >= 100 || value >= 100 ? 100 : 
                    (value == -1 || gridValue == -1 ? -1 : 0);
                    // minValue = (*this)[nxy * ratio + innerXY] > 0 ? minValue : (*this)[nxy * ratio + innerXY]);
                }
            }
            newGridMap[nxy] = value;
        }
    }
    return newGridMap;
}

Indices GridMap::toGridIndices(const Position &inPos) const
{
    return (inPos - origin) / resolution;
}

Position GridMap::toPosition(const Indices &inIndices) const
{
    return (inIndices * resolution) + origin;
}

void GridMap::operator=(Matrix &&rhs)
{
    for (octave_idx_type row = 0; row < rhs.rows(); row++)
    {
        for (octave_idx_type col = 0; col < rhs.cols(); col++)
        {
            this->grid[col + rhs.cols() * row] = rhs(row, col);
        }
    }
}

nav_msgs::OccupancyGrid GridMap::toOccupancyGrid() const
{
    nav_msgs::OccupancyGrid mapMsg;
    mapMsg.header.frame_id = "map";

    mapMsg.info.resolution = this->resolution;
    mapMsg.info.width = this->dim.x;
    mapMsg.info.height = this->dim.y;

    mapMsg.info.origin.position.x = this->origin.x;
    mapMsg.info.origin.position.y = this->origin.y;

    mapMsg.info.origin.orientation.w = 1.0;
    mapMsg.info.origin.orientation.x = 0.0;
    mapMsg.info.origin.orientation.y = 0.0;
    mapMsg.info.origin.orientation.z = 0.0;

    mapMsg.data.reserve(this->grid.size());
    for (auto value : this->grid)
    {
        mapMsg.data.push_back(value);
    }
    return mapMsg;
}

nav_msgs::OccupancyGrid GridMap::toOccupancyGridMeasurementMap() const
{

    float maxValue = *std::max_element(this->grid.begin(), this->grid.end());
    float minValue = *std::min_element(this->grid.begin(), this->grid.end());

    nav_msgs::OccupancyGrid mapMsg;
    mapMsg.header.frame_id = "map";

    mapMsg.info.resolution = this->resolution;
    mapMsg.info.width = this->dim.x;
    mapMsg.info.height = this->dim.y;

    mapMsg.info.origin.position.x = this->origin.x;
    mapMsg.info.origin.position.y = this->origin.y;

    mapMsg.info.origin.orientation.w = 1.0;
    mapMsg.info.origin.orientation.x = 0.0;
    mapMsg.info.origin.orientation.y = 0.0;
    mapMsg.info.origin.orientation.z = 0.0;

    mapMsg.data.reserve(this->grid.size());
    for (auto value : this->grid)
    {
        mapMsg.data.push_back(((value - minValue) * 100.0f / (maxValue - minValue)));
    }
    return mapMsg;
}

MeasurementMap GridMap::toMeasurementMap() const
{

    float maxValue = *std::max_element(this->grid.begin(), grid.end());
    float minValue = *std::min_element(grid.begin(), grid.end());

    if (abs(maxValue - minValue) < 0.5f)
    {
        maxValue = (maxValue + minValue) / 2.0f + 0.5;
        minValue = (maxValue + minValue) / 2.0f - 0.5f;
    }

    MeasurementMap measurementMapMsg;
    measurementMapMsg.header.stamp = ros::Time::now();
    measurementMapMsg.header.frame_id = "map";

    measurementMapMsg.info.resolution = resolution;
    measurementMapMsg.info.width = static_cast<uint32_t>(dim.x);
    measurementMapMsg.info.height = static_cast<uint32_t>(dim.y);
    measurementMapMsg.maxValue = maxValue;
    measurementMapMsg.minValue = minValue;
    measurementMapMsg.type = "temp";

    measurementMapMsg.info.origin.position.x = this->origin.x;
    measurementMapMsg.info.origin.position.y = this->origin.y;

    measurementMapMsg.info.origin.orientation.w = 1.0;
    measurementMapMsg.info.origin.orientation.x = 0.0;
    measurementMapMsg.info.origin.orientation.y = 0.0;
    measurementMapMsg.info.origin.orientation.z = 0.0;

    measurementMapMsg.data.reserve(grid.size());
    for (auto value : grid)
    {
        measurementMapMsg.data.push_back(((value - minValue) * 100.0 / (maxValue - minValue)));
    }

    return measurementMapMsg;
}
}
