#include <thd_mapping/learned_weights.h>
#include <cassert>
#include <iostream>
#include <ros/ros.h>

namespace thd_mapping
{

std::vector<float> createGaussianKernel2(int kernelSize, float sigma = 5.0f)
{

    assert(kernelSize % 2 == 1); //kernelSize is odd
    std::vector<float> kernel(kernelSize * kernelSize);

    float s = 2.0 * sigma * sigma;
    float sum = 0.0;
    int halfKernelSize = (kernelSize / 2);

    float maxValue = 0;
    glm::ivec2 xy;
    for (xy.y = -halfKernelSize; xy.y <= halfKernelSize; xy.y++)
    {
        for (xy.x = -halfKernelSize; xy.x <= halfKernelSize; xy.x++)
        {
            float r = glm::length(glm::vec2(xy));
            float &kernelValue = kernel[(xy.x + halfKernelSize) + (xy.y + halfKernelSize) * kernelSize];
            kernelValue = (exp(-(r * r) / s)) / (glm::pi<float>() * s);
            //   sum += kernelValue;
            maxValue = glm::max(kernelValue, maxValue);
        }
    }
    for (int i = 0; i < kernelSize; ++i)
        for (int j = 0; j < kernelSize; ++j)
            kernel[j + i * kernelSize] /= maxValue;

    return kernel;
}

LearnedWeights::LearnedWeights(GridMap &&grid, std::vector<DataPoint> sensorDataPoints, int kernelSize)
    : GDMAlgorithm(std::move(grid)),
      mSensorDataPoints(sensorDataPoints),
      mWeights(sensorDataPoints.size()),
      mKernel(createGaussianKernel2(kernelSize, 6.0f)),
      mKernelSize(kernelSize)
{

    for (int i = 0; i < mSensorDataPoints.size(); i++)
    {
        mWeights[i] = mGrid;
        glm::ivec2 xy;
        for (xy.y = 0; xy.y < mGrid.dim.y; xy.y++)
        {
            for (xy.x = 0; xy.x < mGrid.dim.x; xy.x++)
            {

                float sumDist = 0;

                for (auto &dataPoint : mSensorDataPoints)
                {
                    glm::vec2 gridPos = mWeights[i].toGridIndices(dataPoint.position);
                    sumDist += glm::length2(glm::vec2(xy) - mWeights[i].toGridIndices(dataPoint.position));
                }
                float avgDist = sumDist / mSensorDataPoints.size();
                float contribution = 0;

                float contributionFactor = 1.0f - glm::length2(glm::vec2(xy) - mWeights[i].toGridIndices(mSensorDataPoints[i].position)) / sumDist;

                // size_t gridIndex = size_t(xy.x + xy.y * mWeights[i].dim.y);
                float &weight = mWeights[i][xy];
                weight = contributionFactor;
            }
        }
    }
}

void LearnedWeights::addDataPoints(const std::vector<DataPoint> &dataPoints)
{
    mSensorDataPoints = dataPoints;
}

void LearnedWeights::addDataPoint(const DataPoint &dataPoint)
{
    mMobileDataPoint = dataPoint;
    mHasNewData = true;
}

void LearnedWeights::updateMap()
{
    if (mHasNewData)
    {
        mHasNewData = false;
        glm::vec2 gridMapPos = this->mGrid.toGridIndices(glm::vec2(mMobileDataPoint.position.x, mMobileDataPoint.position.y));
        float gamma = 1.0f;

        int halfKernelSize = (mKernelSize / 2);
        glm::ivec2 xy;
        for (xy.y = -halfKernelSize; xy.y <= halfKernelSize; xy.y++)
        {
            for (xy.x = -halfKernelSize; xy.x <= halfKernelSize; xy.x++)
            {

                glm::ivec2 gridPos = {gridMapPos.x + xy.x, gridMapPos.y + xy.y};
                if (!(gridPos.x >= 0 && gridPos.x < this->mGrid.dim.x && gridPos.y >= 0 && gridPos.y < this->mGrid.dim.y))
                    continue;

                float prediction = 0;
                float bestFit = std::numeric_limits<float>::infinity();
                for (int i = 0; i < mSensorDataPoints.size(); i++)
                {
                    prediction += mWeights[i][gridPos] * mSensorDataPoints[i].data;
                    bestFit = glm::min(bestFit, glm::abs(mSensorDataPoints[i].data - mMobileDataPoint.data));
                }
                float predictionError = prediction - mMobileDataPoint.data;

                float totalWeightValues = 0;
                for (int i = 0; i < mSensorDataPoints.size(); i++)
                {
                    float &weight = mWeights[i][gridPos];
                    float sensorError = mSensorDataPoints[i].data - mMobileDataPoint.data;

                    float sensorContribution = abs(sensorError);

                    // float correction = 1.0f / (1.0f + glm::exp(-sensorContribution));
                    float correction = 1.0f / glm::max(sensorContribution - bestFit, 0.01f);

                    if (xy.x == 0 && xy.y == 0)
                        ROS_INFO("Pe:%f,SC:%f,C:%f,weight:%f\n", bestFit, sensorContribution, correction, weight);

                    // float kernelValue = 0.5f;
                    const float kernelValue = mKernel[(xy.x + halfKernelSize) + (xy.y + halfKernelSize) * mKernelSize];
                    weight = weight * (1.0f - kernelValue * gamma) + correction * kernelValue * gamma;

                    weight = glm::clamp(weight, 0.0f, 1.0f);
                    // // weight = correctness ;
                    totalWeightValues += glm::abs(weight);
                }
                for (int i = 0; i < mSensorDataPoints.size(); i++) //normalize weights
                {
                    mWeights[i][gridPos] *= 1.0f / totalWeightValues;
                }
            }
        }

    }

    glm::ivec2 xy;
    for (xy.y = 0; xy.y < mGrid.dim.y; xy.y++)
    {
        for (xy.x = 0; xy.x < mGrid.dim.x; xy.x++)
        {
            float &cellValue = mGrid[xy];
            cellValue = 0;
            for (int i = 0; i < mSensorDataPoints.size(); i++) //normalize weights
            {
                cellValue += mWeights[i][xy] * mSensorDataPoints[i].data;
            }
        }
    }
}
}
