// std
#include <memory>
#include <string>

// ROS core
#include <ros/ros.h>
#include <ros/time.h>

// msgs
#include <nav_msgs/OccupancyGrid.h>
#include <geometry_msgs/PoseStamped.h>
#include <sensor_msgs/Image.h>
#include <actionlib_msgs/GoalStatusArray.h>
#include <move_base_msgs/MoveBaseActionGoal.h>

#include <glm/vec3.hpp>

// util
#include <thd_mapping/MeasurementMap.h>
#include <thd_mapping/oct_runner.hpp>
#include <thd_mapping/util.h>

// ALGS
#include <thd_mapping/gaussian_kernel_weight_mapper.h>
#include <thd_mapping/grf_algorithm.hpp>
#include <thd_mapping/kernel_dmv_algorithm.hpp>
#include <thd_mapping/weighted_average_mapper.h>

// Nodes
#include <thd_mapping/basic_node.h>
#include <thd_mapping/median_node.hpp>
#include <thd_mapping/sensor_node.h>
#include <thd_mapping/smoothing_node.hpp>

// goal finding
#include <thd_mapping/goal_finding_algorithm.hpp>
#include <thd_mapping/max_variance_goal_finder.hpp>

#define PASTE(a) a ## Ptr
#define PASTE2(a) PASTE(a)
#define ALGORITHM_PTR PASTE2( ALGORITHM_TYPE )
#define NODE_PTR PASTE2( NODE_TYPE )
#define NAV_PTR PASTE2( NAV_TYPE )
#define CTOR_EXTRA


#define GRF 1
#define WAM 2
#define GKW 3
#define KDM 4
#define LWA 5
#define ALG GRF

#if ALG == GRF
#define ALGORITHM_TYPE thd_mapping::GaussianRandomFieldAlgorithm
#elif ALG==WAM
#define ALGORITHM_TYPE thd_mapping::WeightedAverageMapper
#elif ALG==GKW
#define ALGORITHM_TYPE thd_mapping::GaussianKernelWeightMapper
#define CTOR_EXTRA , 9
#elif ALG==KDM
#define ALGORITHM_TYPE thd_mapping::KernelDMVAlgorithm
#define CTOR_EXTRA , 0.5, 1.0, true
#elif ALG==LWA
#define ALGORITHM_TYPE thd_mapping::LearnedWeights
#endif

#define NODE_TYPE thd_mapping::BasicNode
#define NODE_EXTRA_ARGS

// Not used
#define NAV_TYPE thd_mapping::MaxVarianceGoalFinder


#define DEBUG_PRINT(x) ROS_INFO_STREAM(x)
//#define DEBUG_PRINT(x)


void setupNavigationAlgorithms(ros::NodeHandle& n);
void setupPublishers(ros::NodeHandle& n);
void setupSensorNodes();
void setupAlgorithms();
void publishMaps();
ros::Publisher MeanTempMapPublisher, MeanTempMeasurementMapPublisher;
ros::Publisher MeanHumidityMapPublisher, MeanHumidityMeasurementMapPublisher;
ros::Publisher MeanDustMapPublisher, MeanDustMeasurementMapPublisher;
ros::Publisher VarianceHumidityMapPublisher, VarianceHumidityMeasurementMapPublisher;
ros::Publisher VarianceDustMapPublisher, VarianceDustMeasurementMapPublisher;
ros::Publisher VarianceTempMapPublisher, VarianceTempMeasurementMapPublisher;


ros::Publisher MapScaledPublisher;
ros::Publisher MaxVarianceGoalPublisher;

thd_mapping::MaxVarianceGoalFinderPtr MaxVarianceNode;
ros::Subscriber OccupancyMapSubscriber;
ros::Subscriber GoalReachedSubscriber;

std::vector <thd_mapping::GDMAlgorithmPtr> Algorithms;
std::vector <thd_mapping::SensorNodePtr> SensorNodes;
thd_mapping::GridMap OccupancyMap;


bool HasGridMap = false;


void positionCallback(const geometry_msgs::PoseStamped::ConstPtr& msg)
{
    MaxVarianceNode->setPosition({(float)msg->pose.position.x,(float)msg->pose.position.y});
}

void goalCallback(const actionlib_msgs::GoalStatusArray::ConstPtr& msg)
{
    bool hasActiveGoal = false;
    for (auto goal : msg->status_list)
    {
        if (goal.status == goal.PENDING || goal.status == goal.ACTIVE || goal.status == goal.RECALLING || goal.status == goal.PREEMPTING)
        {
            hasActiveGoal = true;
            // DEBUG_PRINT("hasActiveGoal == true");
        }
    }
    if (!hasActiveGoal)
    {
        // DEBUG_PRINT("hasActiveGoal == false");
        move_base_msgs::MoveBaseActionGoal goalmsg;
        goalmsg.header.stamp = ros::Time::now();
        goalmsg.header.frame_id = "map";

        
        // thd_mapping::Position new_goal = OccupancyMap.toPosition(MaxVarianceNode->getGoal(Algorithms[0]->getVarianceMap(),OccupancyMap));
        // new_goal.y -= 0.5f*OccupancyMap.resolution;
        // new_goal.x += 0.5f*OccupancyMap.resolution;
        thd_mapping::Position new_goal = MaxVarianceNode->getGoal(Algorithms[0]->getVarianceMap(),OccupancyMap);
        // new_goal.y -= 0.5f*OccupancyMap.resolution;
        // new_goal.x += 0.5f*OccupancyMap.resolution;
        goalmsg.goal.target_pose.header.frame_id = "map";
        goalmsg.goal.target_pose.pose.position.x = new_goal.x;
        goalmsg.goal.target_pose.pose.position.y = new_goal.y;
        goalmsg.goal.target_pose.pose.orientation.w = 1.0f;
        MaxVarianceGoalPublisher.publish(goalmsg);
    }
}

void mapCallback(const nav_msgs::OccupancyGrid::ConstPtr& msg)
{
    DEBUG_PRINT("Recieved occupancy map");
    glm::ivec2 lDim = {
        lDim.x = msg->info.width,
        lDim.y = msg->info.height
    };
    
    OccupancyMap = thd_mapping::GridMap(
        std::vector<float>(&msg->data[0], &msg->data[0] + msg->data.size()),
        // std::move(grid),
        lDim,
        glm::vec2(msg->info.origin.position.x, msg->info.origin.position.y),
        msg->info.resolution);
    OccupancyMap = OccupancyMap.downSample(5);
    HasGridMap = true;
}

int main(int argc, char *argv[])
{
    ros::init(argc, argv, "gas_mapper_main");

    ros::NodeHandle n;
    ros::Rate r(1);

    setupPublishers(n);

    OccupancyMapSubscriber = n.subscribe("/map", 1, &mapCallback);
    
    while (n.ok() && !HasGridMap)
    {
        DEBUG_PRINT("Waiting for /map");
        ros::spinOnce();
        r.sleep();
    }

    setupAlgorithms();

    setupSensorNodes();
    
    setupNavigationAlgorithms(n);

    while (n.ok())
    {
        ROS_INFO("In ros loop");
        ros::spinOnce();
        publishMaps();
        r.sleep();
    }
}


void publishMaps()
{
    DEBUG_PRINT("Publish maps -- ");

    MapScaledPublisher.publish(OccupancyMap.toOccupancyGrid());

    MeanDustMapPublisher.publish(Algorithms[2]->getMap().toOccupancyGridMeasurementMap());
    MeanDustMeasurementMapPublisher.publish(Algorithms[2]->getMap().toMeasurementMap());

    MeanHumidityMapPublisher.publish(Algorithms[1]->getMap().toOccupancyGridMeasurementMap());
    MeanHumidityMeasurementMapPublisher.publish(Algorithms[1]->getMap().toMeasurementMap());

    MeanTempMapPublisher.publish(Algorithms[0]->getMap().toOccupancyGridMeasurementMap());
    MeanTempMeasurementMapPublisher.publish(Algorithms[0]->getMap().toMeasurementMap());

    VarianceDustMapPublisher.publish(Algorithms[2]->getVarianceMap().toOccupancyGridMeasurementMap());
    VarianceDustMeasurementMapPublisher.publish(Algorithms[2]->getVarianceMap().toMeasurementMap());

    VarianceHumidityMapPublisher.publish(Algorithms[1]->getVarianceMap().toOccupancyGridMeasurementMap());
    VarianceHumidityMeasurementMapPublisher.publish(Algorithms[1]->getVarianceMap().toMeasurementMap());

    VarianceTempMapPublisher.publish(Algorithms[0]->getVarianceMap().toOccupancyGridMeasurementMap());
    VarianceTempMeasurementMapPublisher.publish(Algorithms[0]->getVarianceMap().toMeasurementMap());

    DEBUG_PRINT("Finished publishing maps --");
}

void setupAlgorithms()
{
    DEBUG_PRINT("Setting up algorithms -- ");
    // int resolutionFactor = 1;
    glm::ivec2 dim = OccupancyMap.dim;
    float initTemp = 15.0f;
    float initVariance = 0.0;
    float initHumidity = 30.0f;
    float initDust = 20.0f;

    glm::vec2 origin = OccupancyMap.origin;
    
    Algorithms = {
        std::make_shared<ALGORITHM_TYPE>(
            thd_mapping::GridMap(std::vector<float>(dim.x*dim.y, initTemp),
                                 dim,
                                 origin,
                                 OccupancyMap.resolution) CTOR_EXTRA),
        std::make_shared<ALGORITHM_TYPE>(
            thd_mapping::GridMap(std::vector<float>(dim.x*dim.y, initHumidity),
                                 dim,
                                 origin,
                                 OccupancyMap.resolution) CTOR_EXTRA),
        
        std::make_shared<ALGORITHM_TYPE>(
            thd_mapping::GridMap(std::vector<float>(dim.x*dim.y, initDust),
                                 dim,
                                 origin,
                                 OccupancyMap.resolution) CTOR_EXTRA)};

#if ALG==GRF
    for (auto alg : Algorithms )
    {
        static_cast<thd_mapping::GaussianRandomFieldAlgorithm*>(alg.get())->setOccupancyMap(OccupancyMap);
        static_cast<thd_mapping::GaussianRandomFieldAlgorithm*>(alg.get())->setDelay(5);   
    }
#endif
    
    DEBUG_PRINT("Finished etting up algorithms -- ");
}

void setupSensorNodes()
{
    DEBUG_PRINT("Setting up sensor nodes --");
    boost::function<void(const thd_mapping::SensorNode&)> func5 = boost::bind(&thd_mapping::dataPointCallback, _1, Algorithms, 0);
    boost::function<void(const thd_mapping::SensorNode&)> func9 = boost::bind(&thd_mapping::dataPointCallback, _1, Algorithms, 1);
    boost::function<void(const thd_mapping::SensorNode&)> funcusb = boost::bind(&thd_mapping::dataPointCallback, _1, Algorithms, 2);

    SensorNodes = {
        std::make_shared<NODE_TYPE>(std::string("n5"), thd_mapping::Position(3.72,4.0), func5 NODE_EXTRA_ARGS ),
        std::make_shared<NODE_TYPE>(std::string("n9"), thd_mapping::Position(1.25,-3.96), func9 NODE_EXTRA_ARGS ),
        std::make_shared<NODE_TYPE>(std::string("usb"), thd_mapping::Position(0,5), funcusb NODE_EXTRA_ARGS )
    };
    
    DEBUG_PRINT("Finished setting up sensor nodes --");
}

void setupPublishers(ros::NodeHandle& n)
{
    DEBUG_PRINT("Setting up publishers --");    

    

    // FOR ROS, Temp Hum Dust
    MeanTempMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/meantemp", 1000);
    MeanHumidityMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/meanhumidity", 1000);
    MeanDustMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/meandust", 1000);

    // For Web, Temp Hum Dust
    MeanTempMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/meantempmeasurement", 1000);
    MeanHumidityMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/meanhumiditymeasurement", 1000);
    MeanDustMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/meandustmeasurement", 1000);

    // For Ros, Variance Temp Hum Dust
    VarianceTempMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/variancetemp", 1000);
    VarianceHumidityMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/variancehumidity", 1000);
    VarianceDustMapPublisher= n.advertise<nav_msgs::OccupancyGrid>("/variancedust", 1000);

    // For Web, Variance Temp Hum Dust
    VarianceTempMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/variancetempmeasurement", 1000);
    VarianceHumidityMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/variancehumiditymeasurement", 1000);
    VarianceDustMeasurementMapPublisher= n.advertise<thd_mapping::MeasurementMap>("/variancedustmeasurement", 1000);

    DEBUG_PRINT("Finished setting up publishers --");    
}

void setupNavigationAlgorithms(ros::NodeHandle& n)
{
    // MaxVarianceNode = std::make_shared<thd_mapping::MaxVarianceGoalFinder>(Algorithms[0]->getVarianceMap(),OccupancyMap);
    MaxVarianceNode = std::make_shared<thd_mapping::MaxVarianceGoalFinder>();
    

    GoalReachedSubscriber = n.subscribe("/move_base/status", 1, &goalCallback);
    MaxVarianceGoalPublisher = n.advertise<move_base_msgs::MoveBaseActionGoal>("/maxvariancegoal", 1000);
    MapScaledPublisher= n.advertise<nav_msgs::OccupancyGrid>("/mapScaled", 1000);
}
