#include <ros/ros.h>
#include <ros/time.h>
#include <nav_msgs/OccupancyGrid.h>
#include <string>

#include <string>
#include <thd_mapping/sensor_node.h>
#include <thd_mapping/basic_node.h>
#include <glm/vec3.hpp>
#include <memory>

#include "thd_mapping/util.h"
#include "thd_mapping/learned_weights.h"

#define TEST_MAPPING_IMAGE
ros::Publisher meanMapPub;
ros::Publisher varianceMapPub;
ros::Publisher measurementMapPub;
ros::Subscriber mapSub;

thd_mapping::GridMap gMap;
bool ghasGridMap = false;
std::map<int, thd_mapping::DataPoint> gSensorDataPoints;

void publishMaps(const thd_mapping::GridMap &meanMap, const thd_mapping::GridMap &varianceMap);

void sensorDataPointCallback(const thd_mapping::SensorNode &node,
                             const std::vector<thd_mapping::GDMAlgorithmPtr> &mapping_alg_ptr, int id)
{
    thd_mapping::Position p = node.getPosition();
    thd_mapping::Reading r = node.getReading();

    thd_mapping::DataPoint newDataPoint;
    newDataPoint.data = r.x;
    newDataPoint.position = p;
    newDataPoint.node_id = id;

    gSensorDataPoints[id] = newDataPoint;
}

void mapCallback(const nav_msgs::OccupancyGrid::ConstPtr &msg)
{
    ROS_INFO("Received Map.");
    glm::ivec2 lDim = {
        lDim.x = msg->info.width,
        lDim.y = msg->info.height};

    gMap = thd_mapping::GridMap(
        std::vector<float>(&msg->data[0], &msg->data[0] + msg->data.size()),
        // std::move(grid),
        lDim,
        glm::vec2(msg->info.origin.position.x, msg->info.origin.position.y),
        msg->info.resolution);
    ghasGridMap = true;
}

int main(int argc, char *argv[])
{
    ros::init(argc, argv, "test_grf_node");
    ros::NodeHandle n;
    ros::Rate r(1);
    meanMapPub = n.advertise<nav_msgs::OccupancyGrid>("/mean_map", 5);
    measurementMapPub = n.advertise<thd_mapping::MeasurementMap>("measurement_map", 5);
    mapSub = n.subscribe("/map", 1, &mapCallback);

    while (n.ok() && !ghasGridMap)
    {
        ROS_INFO("Waiting for /map");
        ros::spinOnce();
        r.sleep();
    }

    int resolutionFactor = 1;
    glm::ivec2 dim = gMap.dim / resolutionFactor;
    float initTemp = 15.0f;
    float initVariance = 0.0;
    float initHumidity = 30.0f;
    float initDust = 20.0f;
    int kernelSize = 61;

    std::vector<thd_mapping::GDMAlgorithmPtr> algorithms;

    boost::function<void(const thd_mapping::SensorNode &)> func5 = boost::bind(&sensorDataPointCallback, _1, boost::cref(algorithms), 0);
    boost::function<void(const thd_mapping::SensorNode &)> func9 = boost::bind(&sensorDataPointCallback, _1, boost::cref(algorithms), 1);
    boost::function<void(const thd_mapping::SensorNode &)> funcusb = boost::bind(&thd_mapping::dataPointCallback, _1, boost::cref(algorithms), 2);

    std::vector<std::shared_ptr<thd_mapping::SensorNode>> nodes = {
        std::make_shared<thd_mapping::BasicNode>(std::string("n5"), thd_mapping::Position(3.72, 4.0), func5),
        std::make_shared<thd_mapping::BasicNode>(std::string("n9"), thd_mapping::Position(1.25, -3.96), func9),
        std::make_shared<thd_mapping::BasicNode>(std::string("usb"), thd_mapping::Position(0, 0), funcusb)};

    while (n.ok() && gSensorDataPoints.size() < 2)
    {
        ROS_INFO("Waiting static sensorDataPoints");
        ros::spinOnce();
        r.sleep();
    }

    std::vector<thd_mapping::DataPoint> sensorDataPoints;
    for (auto &dataPointPair : gSensorDataPoints)
    {
        sensorDataPoints.push_back(dataPointPair.second);
    }

    algorithms.push_back(std::make_shared<thd_mapping::LearnedWeights>(
        thd_mapping::GridMap(std::vector<float>(dim.x * dim.y, initTemp), dim, {-5, -5}, gMap.resolution * resolutionFactor),
        sensorDataPoints,
        kernelSize));
    sensorDataPoints.clear();

    while (n.ok())
    {
        ROS_INFO("In ros loop");
        ros::spinOnce();

        for (auto &dataPointPair : gSensorDataPoints)
        {
            sensorDataPoints.push_back(dataPointPair.second);
        }
        algorithms[0]->addDataPoints(sensorDataPoints);
        sensorDataPoints.clear();


        std::static_pointer_cast<thd_mapping::LearnedWeights>(algorithms[0])->updateMap();

        meanMapPub.publish(algorithms[0]->getMap().toOccupancyGrid());
        measurementMapPub.publish(algorithms[0]->getMap().toMeasurementMap());
        r.sleep();
    }
}
